#include <math.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "axml.h"

extern char run_id[40], workdir[2048], *myTreestr;
extern int treeStringLength;
extern rawdata *rdtaSupertree;

char *treeStringStBuilder (char *treestr, tree *tr, nodeptr p, int form)
    /* write string with representation of tree */
    /*   form == 1 -> Newick tree */
    /*   form == 2 -> Prolog fact */
    /*   form == 3 -> PHYLIP tree */
  { /* treeString */
    double  x, z;
    char  *nameptr;
    int    c;
   
    if (p->tip) {
      char str[20];
      sprintf(str, "%d", p->number);
      nameptr = str;/*p->name;*/
      while (c = *nameptr++) 
	{
	  if (form != treePHYLIP) {if (c == '\'') *treestr++ = '\'';}
	  else if (c == ' ') {c = '_';}
	  *treestr++ = c;
        }
    }

    else {
      *treestr++ = '(';
      treestr = treeStringStBuilder(treestr, tr, p->next->back, form);
      *treestr++ = ',';
      treestr = treeStringStBuilder(treestr, tr, p->next->next->back, form);
      if (p == tr->start->back) {
        *treestr++ = ',';
        treestr = treeStringStBuilder(treestr, tr, p->back, form);
        }
      *treestr++ = ')';
    }

    if (p == tr->start->back) {
      (void) sprintf(treestr, ";\n");
      }
    else {
      z = p->z;
      if (z < zmin) z = zmin;
#ifdef REVERSIBLE
     x = -log(z) * tr->rdta->fracchangeav; 
#else
      x = -log(z) * tr->rdta->fracchange;
#endif
      (void) sprintf(treestr, ": %8.6f", x);  /* prolog needs the space */
      }

    while (*treestr) treestr++;     /* move pointer up to null termination */
    return  treestr;
  } /* treeString */




#ifdef DEBUG
char *treeStringDivide (char *treestr, tree *tr, nodeptr p, int form)
    /* write string with representation of tree */
    /*   form == 1 -> Newick tree */
    /*   form == 2 -> Prolog fact */
    /*   form == 3 -> PHYLIP tree */
  { /* treeString */
    double  x, z;
    char  *nameptr;
    int    c;

    if (p == tr->start->back) {
      if (form != treePHYLIP) {
        if (form == treeProlog) {
          (void) sprintf(treestr, "phylip_tree(");
          while (*treestr) treestr++;            /* move pointer to null */
          }

        (void) sprintf(treestr, "[&&%s: version = '%s'",
                                 programName, programVersion);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %15.13g",
                                 likelihood_key, tr->likelihood);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %d", ntaxa_key, tr->ntips);
        while (*treestr) treestr++;

        (void) sprintf(treestr,", %s = %d", opt_level_key, tr->opt_level);
        while (*treestr) treestr++;

        (void) sprintf(treestr, ", %s = %d", smoothed_key, tr->smoothed);
        while (*treestr) treestr++;

        (void) sprintf(treestr, "]%s", form == treeProlog ? ", " : " ");
        while (*treestr) treestr++;
        }
    }

    if (p->tip) {
      char str[20], str2[5];
      strcpy(str, p->name);
      strcat(str, "__");
      sprintf(str2, "%d", p->set);
      strcat(str, str2);
      if (form != treePHYLIP) *treestr++ = '\'';
      nameptr = str;/*p->name;*/
      while (c = *nameptr++) {
        if (form != treePHYLIP) {if (c == '\'') *treestr++ = '\'';}
        else if (c == ' ') {c = '_';}
        *treestr++ = c;
        }
      if (form != treePHYLIP) *treestr++ = '\'';
    }

    else {
      *treestr++ = '(';
      treestr = treeStringDivide(treestr, tr, p->next->back, form);
      *treestr++ = ',';
      treestr = treeStringDivide(treestr, tr, p->next->next->back, form);
      if (p == tr->start->back) {
        *treestr++ = ',';
        treestr = treeStringDivide(treestr, tr, p->back, form);
        }
      *treestr++ = ')';
    }

    if (p == tr->start->back) {
      (void) sprintf(treestr, ":0.0%s\n", (form != treeProlog) ? ";" : ").");
      }
    else {
      z = p->z;
      if (z < zmin) z = zmin;
#ifdef REVERSIBLE
     x = -log(z) * tr->rdta->fracchangeav; 
#else
      x = -log(z) * tr->rdta->fracchange;
#endif
      (void) sprintf(treestr, ": %8.6f", x);  /* prolog needs the space */
      }

    while (*treestr) treestr++;     /* move pointer up to null termination */
    return  treestr;
  } /* treeString */


char *treeStringDivide2 (char *treestr, tree *tr, nodeptr p, int form)
    /* write string with representation of tree */
    /*   form == 1 -> Newick tree */
    /*   form == 2 -> Prolog fact */
    /*   form == 3 -> PHYLIP tree */
  { /* treeString */
    double  x, z;
    char  *nameptr;
    int    c;

   
    if (p->tip) {
      char str[20], str2[5];
      strcpy(str, p->name);
      /*strcat(str, "__");
      sprintf(str2, "%d", p->set);
      strcat(str, str2);*/
      if (form != treePHYLIP) *treestr++ = '\'';
      nameptr = str;/*p->name;*/
      while (c = *nameptr++) {
        if (form != treePHYLIP) {if (c == '\'') *treestr++ = '\'';}
        else if (c == ' ') {c = '_';}
        *treestr++ = c;
        }
      if (form != treePHYLIP) *treestr++ = '\'';
    }

    else {
      *treestr++ = '(';
      treestr = treeStringDivide2(treestr, tr, p->next->back, form);
      *treestr++ = ',';
      treestr = treeStringDivide2(treestr, tr, p->next->next->back, form);     
      *treestr++ = ')';
    }

   
      z = p->z;
      if (z < zmin) z = zmin;
#ifdef REVERSIBLE
     x = -log(z) * tr->rdta->fracchangeav; 
#else
      x = -log(z) * tr->rdta->fracchange;
#endif
      (void) sprintf(treestr, ": %8.6f", x);  /* prolog needs the space */
      

    while (*treestr) treestr++;     /* move pointer up to null termination */
    return  treestr;
  } /* treeString */
#endif


static int doublecompare(const void *p1, const void *p2)
{
  double i = *((double *)p1);
  double j = *((double *)p2);
  
  if (i > j)
    return (1);
  if (i < j)
    return (-1);
  return (0);
}
int calcRec(nodeptr p, tree *tr)
{
  int acc = 0;
  nodeptr q;

  if(p->tip) 
    return 1;
  else
    {
     q = p->next;
     while (q != p) 
       {
	 acc += calcRec(q->back, tr);
	 q = q->next;
       } 
    }

  return acc;
}

int calcWeights(nodeptr p, tree *tr)
{
  int left, right = 0, diff;
  nodeptr q;      
  left = calcRec(p->back, tr);
  if(!p->tip) 
    {
      q = p->next;
      while(q != p)
	{
	  right += calcRec(q->back, tr);
	  q = q->next;
	}
    }

  if(left >= right)
    {
      diff = left - right;
      /*fprintf(stderr,"%d\n", left - right);*/
    }
  else
    {
      diff = right - left;
      /*fprintf(stderr,"%d\n", right - left);*/
    }

  return diff;
}

int calcRecSet(nodeptr p, tree *tr, int set)
{
  int acc = 0;
  nodeptr q;

  if(p->tip) 
    return ((p->set == set) ? 1 : 0);
  else
    {
     q = p->next;
     while (q != p) 
       {
	 acc += calcRecSet(q->back, tr, set);
	 q = q->next;
       } 
    }

  return acc;
}



int calcWeightsSet(nodeptr p, tree *tr, int set)
{
  int left, right = 0, diff;
  nodeptr q;      
  left = calcRecSet(p->back, tr, set);
  if(!p->tip) 
    {
      q = p->next;
      while(q != p)
	{
	  right += calcRecSet(q->back, tr, set);
	  q = q->next;
	}
    }

  return ((left >= right) ? (left - right) : (right -left));
}

void dRec(nodeptr p)
{
  nodeptr q;

  if(p->tip) 
    fprintf(stderr,"%s\n", p->name);
  else
    {
     q = p->next;
     while (q != p) 
       {
	 dRec(q->back);
	 q = q->next;
       } 
    }

}


void displayTaxa(nodeptr p)
{
  nodeptr q;      
  fprintf(stderr,"LEFT:\n");
  dRec(p->back);
  fprintf(stderr,"RIGHT:\n");
  if(!p->tip) 
    {
      q = p->next;
      while(q != p)
	{
	  dRec(q->back);
	  q = q->next;
	}
    }
}

double accDist = 0;

void distanceRec(nodeptr p, double d, int set)
{
  nodeptr q;
  double f;

  if(p->tip) 
    {
      p->distance = p->z + d;
      p->set = set;
      /*fprintf(stderr,"tip %s %f %d\n", p->name, p->distance, p->set);*/
      accDist += p->distance;
    }
  else
    { 
     q = p->next;    
     while (q != p) 
       {
	 q->distance = q->z + d;
	 f = q->distance;
	 distanceRec(q->back, f, set);
	 q = q->next;
       } 
    }  
}

double tipDistance(nodeptr p)
{
  nodeptr q;
  double left, right;    

  accDist = 0;
  
  distanceRec(p->back, 0, 0);
  left = accDist;

  
  accDist = 0;
  if(!p->tip) 
    {
      q = p->next;
      while(q != p)
	{
	  distanceRec(q->back, 0, 1);
	  q = q->next;
	}
    }

  right = accDist;

  return (left <= right? (right - left): (left - right));
}









void treeDivisor(tree *tr)
{
  int i, max_w = 0, w, wprev = tr->mxtips, k, max_d, s0, s1;
  double d, dprev = 20000000000000, threshold;
  double *set0, *set1;
  for(i = tr->mxtips + 1; i < (2 * tr->mxtips - 1); i++)
    {      

      for(k = 1; k < (2 * tr->mxtips - 1); k++)
	tr->nodep[i]->distance = 0;
      w = calcWeights(tr->nodep[i], tr);
      d = tipDistance(tr->nodep[i]);
      if(w < wprev)
	{
	  wprev = w;
	  max_w = i;
	}
      if(d < dprev)
	{
	  dprev = d;
	  max_d = i;
	}
      /*fprintf(stderr,"I: %d W %d D %f\n", i, w, d);*/
    }

  fprintf(stderr,"D %f i_d %d, W %d i_w %d\n",dprev, max_d, wprev, max_w);
  
  if(max_d != max_w)
    fprintf(stderr,"We have got a problem\n");
  tipDistance(tr->nodep[max_w]);
  
  
  set0 = (double *)malloc(2 * tr->mxtips * sizeof(double));
  set1 = (double *)malloc(2 * tr->mxtips * sizeof(double));

  for(i = 1; i < 2 * tr->mxtips; i++)
    {
      set0[i] = 3000000.0;
      set1[i] = 3000000.0;
    }

  s0 = 0;
  s1 = 0;
  for(i = 1; i <= tr->mxtips; i++)
    {
      if(tr->nodep[i]->set == 0)
	{
	  set0[s0] = tr->nodep[i]->distance;
	  s0++;
	}
      else
	{
	  if(tr->nodep[i]->set == 1)
	    {
	      set1[s1] = tr->nodep[i]->distance;
	      s1++; 
	    }
	  else
	    {
	      fprintf(stderr,"hjgdsa malaka\n");
	      exit(-1);
	    }
	}
    }

  fprintf(stderr,"S0: %d S1: %d\n", s0, s1);
  for(i = 1; i <= tr->mxtips; i++)
    {
      if(tr->nodep[i]->set == 0)
	fprintf(stderr,"S0 %s\n", tr->nodep[i]->name);
      else
	fprintf(stderr,"S1 %s\n", tr->nodep[i]->name);
    }
  exit(1);
  qsort(set0, tr->mxtips, sizeof(double), doublecompare);
  qsort(set1, tr->mxtips, sizeof(double), doublecompare);

  threshold = set1[(int)(s1/10)];
  k = 0;
  for(i = 1; i < tr->mxtips; i++)
    {
      if(tr->nodep[i]->set == 0 || 
	 (tr->nodep[i]->set == 1 && tr->nodep[i]->distance <= threshold))
	{
	  /*	  fprintf(stderr,"%d S0 %s\n", k, tr->nodep[i]->name);*/
	  k++;
	}
    }

  threshold = set0[(int)(s0/10)];
  k = 0;
  for(i = 1; i < tr->mxtips; i++)
    {
      if(tr->nodep[i]->set == 1 || 
	 (tr->nodep[i]->set == 0 && tr->nodep[i]->distance <= threshold))
	{
	  /*  fprintf(stderr,"%d S1 %s\n", k, tr->nodep[i]->name);*/
	  k++;
	}
    }

  /*for(i = tr->mxtips + 1; i < (2 * tr->mxtips - 1); i++)
     {   
       fprintf(stderr,"SET0 %d\n", calcWeightsSet(tr->nodep[i], tr, 0));
     }
   for(i = tr->mxtips + 1; i < (2 * tr->mxtips - 1); i++)
     {   
       fprintf(stderr,"SET1 %d\n", calcWeightsSet(tr->nodep[i], tr, 1));
       }*/

  /*for(i = 0; i < s0 + 1; i++)
    fprintf(stderr,"%f\n", set0[i]);

  for(i = 0; i < s1 + 1; i++)
  fprintf(stderr,"%f\n", set1[i]);*/


  free(set0);
  free(set1);

  /*fprintf(stderr,"FFF %d\n", calcWeights(tr->nodep[max], tr));
  displayTaxa(tr->nodep[max]);
  tipDistance(tr->nodep[max]);*/
}

void distanceGlobalRec(nodeptr p, double d)
{
  nodeptr q;
  double f;

  

  if(p->tip) 
    {
      p->distance = p->z + d;
    }
  else
    { 
     q = p->next;    
     while (q != p) 
       {
	 q->distance = q->z + d;
	 f = q->distance;
	 distanceGlobalRec(q->back, f);
	 q = q->next;
       } 
    }  
}


void distanceGlobal(nodeptr p)
{
   nodeptr q;
   
   distanceGlobalRec(p->back, 0);

   if(!p->tip) 
     {
       q = p->next;
       while(q != p)
	 {
	   distanceGlobalRec(q->back, 0);
	   q = q->next;
	 }
     }

}

void addSequencesSet(tree *tr, int set, int overlap, int divisor, int dC, int *filter, int *count)
{
  double *distances;
  int i, d_count, divisor_tips, j;  
  double threshold;
  int sp_count = 0;
   char inverseMap[21] = {'z', 'A', 'C','M','G','R','S','V', 'T', 'W', 'Y','H','K','D','B','N','O','X','?','-'};
   FILE *outf;
   char outfName[2048], str[16];
   

  distances = (double *)malloc(2 * tr->mxtips * sizeof(double));
  for(i = 1; i < 2 * tr->mxtips; i++)    
    distances[i] = 3000000.0;
  /*divisor_tips = findDivisorTips(tr, set);*/
  divisor_tips = findDivisorDistance(tr, set);
  distanceGlobal(tr->nodep[divisor_tips]);
    /*distanceGlobal(tr->nodep[divisor]);*/
  d_count = 0;
  for(i = 1; i <= tr->mxtips; i++)
    {
      if(tr->nodep[i]->set != set) 
	{
	  distances[d_count] = tr->nodep[i]->distance;
	  d_count++;
	}
    }
  
  qsort(distances, tr->mxtips, sizeof(double), doublecompare);
  threshold = distances[10];

  for(i = 1; i <= tr->mxtips; i++)
    {      
      if((tr->nodep[i]->set != set && tr->nodep[i]->distance <= threshold) || tr->nodep[i]->set == set)
       {	
	 sp_count++;
	 filter[i] = 1;
       }
     else 
       filter[i] = 0;
    }
  count[set] = sp_count;
  strcpy(outfName, workdir);
  strcat(outfName, "/RAxML_subAlignment.");
  strcat(outfName, run_id);
  strcat(outfName, ".");
  sprintf(str, "%d", set);
  strcat(outfName, str);  
  
#ifdef DEBUG
  fprintf(stderr,"set_%d seq: %d\n", set, sp_count);
#endif
  outf = fopen(outfName, "w");
  fprintf(outf, "%d %d\n",sp_count, tr->rdta->sites);
  for(i = 1; i <= tr->mxtips; i++)
    {
      if(tr->nodep[i]->set == set || (tr->nodep[i]->set != set && tr->nodep[i]->distance <= threshold))
	{
	  fprintf(outf,"%s        \0", tr->nodep[i]->name);
	  for(j = 1; j <= tr->rdta->sites; j++)
	    fprintf(outf, "%c", inverseMap[rdtaSupertree->y[i][j]]);
	  fprintf(outf, "\n");
	}
    }
  


  fclose(outf);     
  free(distances);
}

void addSequencesSetAlgo2(tree *tr, int overlap, int **filter, int *count, int center, int numberOfSets)
{
  double **distances;
  int i, j, k;  
  double *thresholds;
  int sp_count = 0;
  int set_count = 0;
   char inverseMap[21] = {'z', 'A', 'C','M','G','R','S','V', 'T', 'W', 'Y','H','K','D','B','N','O','X','?','-'};
   FILE *outf;
   char outfName[2048];
   
   for(i = 1; i < (2 * tr->mxtips - 1); i++)
     tr->nodep[i]->distance = 0;
  
   distanceGlobal(tr->nodep[center]);
   
   distances = (double **)malloc(sizeof(double) * numberOfSets);
   thresholds = (double *)malloc(sizeof(double) * numberOfSets);
   for(i = 0; i < numberOfSets; i++)
     {
       distances[i] = (double *)malloc(sizeof(double) * (tr->mxtips + 1));
       distances[i][0] = 500000000;
       for(j = 1; j <= tr->mxtips; j++)
	 {
	   if(tr->nodep[j]->set == i)
	     distances[i][j] = tr->nodep[j]->distance;
	   else
	     distances[i][j] = 500000000;
	 }
       qsort(distances[i], tr->mxtips, sizeof(double), doublecompare);
       thresholds[i] = distances[i][overlap];
     }
  



  for(i = 0; i < numberOfSets; i++)
    {
      sp_count = 0;
      set_count = 0;
      for(j = 1; j <= tr->mxtips; j++)
	{      
	  if(tr->nodep[j]->set == i)
	    {	
	      sp_count++;
	      set_count++;
	      filter[i][j] = 1;
	    }
	  else 
	    {
	      if(tr->nodep[j]->distance < thresholds[tr->nodep[j]->set])
		{
		  filter[i][j] = 1;
		  sp_count++;
		}
	      else
		filter[i][j] = 0;
	    }
	}
      count[i] = sp_count;
      fprintf(stderr,"set %d with %d species original set %d\n", i, sp_count, set_count);
    }
  
  
  for(i = 0; i < numberOfSets; i++)
    {
      {
	char str[16];
	strcpy(outfName, workdir);
	strcat(outfName, "/RAxML_subAlignment.");
	strcat(outfName, run_id);
	strcat(outfName, ".");
	sprintf(str, "%d", i);
	strcat(outfName, str);  
      }
#ifdef DEBUG
      fprintf(stderr,"set_%d seq: %d\n", set, sp_count);
#endif
      outf = fopen(outfName, "w");
      fprintf(outf, "%d %d\n",count[i], tr->rdta->sites);
      for(j = 1; j <= tr->mxtips; j++)
	{
	  if(filter[i][j])
	    {
	      fprintf(outf,"%s        \0", tr->nodep[j]->name);
	      for(k = 1; k <= tr->rdta->sites; k++)
		fprintf(outf, "%c", inverseMap[rdtaSupertree->y[j][k]]);
	      fprintf(outf, "\n");
	    }
	  
	}      
      fclose(outf);  
      free(distances[i]);
    }
  free(distances);
  free(thresholds);
}


int mpow(int d)
{
  if(d == 0) return 1;
  if(d == 1) return 2;
  return (2 * mpow(d - 1));
}


int findDivisorTips(tree *tr,int set)
{
  int i, max_w = 0, wprev = tr->mxtips, w;

  for(i = tr->mxtips + 1; i < (2 * tr->mxtips - 1); i++)
    {            
      w = calcWeightsSet(tr->nodep[i], tr, set);
      if(w < wprev)
	{
	  wprev = w;
	  max_w = i;
	}
    }

  return max_w;
}
  


double distanceRecSet(nodeptr p, double d, int set)
{
  nodeptr q;
  double f;
  double res;

  if(p->tip) 
    {
      p->distance = p->z + d;
      if(p->set == set)
	accDist += p->distance;
      return p->z;
    }
  else
    { 
     q = p->next;
     res = 0;
     while (q != p) 
       {
	 q->distance = q->z + d;
	 f = q->distance;
	 res += q->z + distanceRecSet(q->back, f, set);
	 q = q->next;
       } 
     return res; 
    }  
}



double tipDistanceSet(nodeptr p, int set)
{
  nodeptr q;
  double left, right;    
  double left2, right2;
  double d1, d2;

  accDist = 0;  
  left2 = distanceRecSet(p->back, 0, set);
  left = accDist;

  
  accDist = 0;
  if(!p->tip) 
    {
      q = p->next;
      while(q != p)
	{
	  right2 = distanceRecSet(q->back, 0, set);
	  q = q->next;
	}
    }

  right = accDist;
  
  d1 =  (left <= right? (right - left): (left - right));
  d2 =  (left2 <= right2? (right2 - left2): (left2 - right2));
	  

  return d1;
}

  
int findDivisorDistance(tree *tr, int set)
{
  int i, k, max_d;
  double d, dprev = 20000000000000000;

  for(i = tr->mxtips + 1; i < (2 * tr->mxtips - 1); i++)
    {      

      for(k = 1; k < (2 * tr->mxtips - 1); k++)
	tr->nodep[i]->distance = 0;
      d = tipDistanceSet(tr->nodep[i], set);
      if(d < dprev)
	{
	  dprev = d;
	  max_d = i;
	}
    } 

  return max_d;
}




void splitTipsRec(nodeptr p, tree *tr, int setl, int setr)
{
  nodeptr q;
  if(p->tip)
    {
      if(p->set == setl) p->set = setr;
    }
  else
    {
     q = p->next;
     while (q != p) 
       {
	 splitTipsRec(q->back, tr, setl, setr);
	 q = q->next;
       } 
    }
}

void splitTips(int sl, int sr, int divisor, tree *tr)
{
  splitTipsRec(tr->nodep[divisor]->back, tr, sl, sr);
}  

int checkRec(nodeptr p, int set)
{
  nodeptr q;
  int result = 0;

  if(p->tip)
    {
      if(p->set == set) return 1;
      else return 0;
    }
  else
    {
     q = p->next;
     while (q != p) 
       {
	 if(!checkRec(q->back, set)) return 0;
	 q = q->next;
       } 
     return 1;
    }
}

#ifdef DEBUG
int check(tree *tr, nodeptr p, int set)
{
  nodeptr q;
  int res;

  res = 0;
  res = checkRec(p->back, set);
  fprintf(stderr,"checking left %d\n", res);

  if(res)
    {
      treeStringDivide2(myTreestr, tr, p->back, 1);
      fprintf(stderr, "%s\n", myTreestr);
    }
  else
    {
      res = 1;
      if(!p->tip) 
	{
	  q = p->next;
	  while(q != p)
	    {
	      res = res && checkRec(q->back, set);
	      q = q->next;
	    }
	}
      fprintf(stderr,"checking right %d\n", res);
      if(res)
	{
	  char *trptr;
	  treeStringDivide2(myTreestr, tr, p, 1);
	  trptr = myTreestr;
	  while(*trptr) trptr++;
	  sprintf(trptr, ";");
	  fprintf(stderr, "%s\n", myTreestr); 
	  str_treeReadLen(myTreestr, tr);
	  fprintf(stderr, "read tree\n");
	  smoothTree(tr, smoothings);
	  evaluate(tr, tr->start);
	  fprintf(stderr, "likelihood %f\n", tr->likelihood);
	  exit(1);
	}
      else
	{
	  fprintf(stderr,"ERROR\n");
	  exit(-1);
	}
    }


}
#endif

void divideRec(tree *tr, int set, int depth, int *splittingNodes)
{
  int divisor, divisor_distance;

  divisor = findDivisorTips(tr, set);
  divisor_distance = findDivisorDistance(tr, set);
  splitTips(set, set + mpow(depth), divisor_distance, tr);
  splittingNodes[set] = divisor_distance;
  splittingNodes[set + mpow(depth)] = divisor_distance;
  /*fprintf(stderr,"depth %d setA %d setB %d Tip_d: %d Dist_d: %d\n", depth, 
    set, set + mpow(depth), divisor, divisor_distance);*/
  if(depth == 0) return;
  divideRec(tr, set, depth - 1, splittingNodes);
  divideRec(tr, set + mpow(depth), depth - 1, splittingNodes);
}

int divideCount = 0;



void divide(tree *tr, int optimizeSubtrees, analdef *adef)
{
  int i, k;
  int *count;
  int descent = 0;
  int *splittingNodes;
  char **names;
  int **filters;
  char *parsimonyTree, *stBuilderTree;
  int allTips;  
  int numberOfSets = adef->sets;
  int overlap;
  int v = adef->sets;
  int centerSplit;
  double perCent = (double)adef->overlap / 100.0;
  FILE *stBuilder, *subtreeFile;
  char stBuilderFileName[2048], subtreeFileName[2048];
 
 
  strcpy(stBuilderFileName,  workdir);
  strcat(stBuilderFileName, "Build.input.");
  strcat(stBuilderFileName, run_id);

  strcpy(subtreeFileName,  workdir);
  strcat(subtreeFileName, "RAxML_subtrees.");
  strcat(subtreeFileName, run_id);


  stBuilder = fopen(stBuilderFileName, "w");

  fprintf(stBuilder,"%d    3\n", tr->mxtips);
  fprintf(stBuilder,"RAxML Tree Divisor. *\n");

  for(i = 1; i <= tr->mxtips; i++)
    fprintf(stBuilder,"%d %s *\n", tr->nodep[i]->number, tr->nodep[i]->name);

  fprintf(stBuilder,"\n");
  fprintf(stBuilder,"%d \n", numberOfSets);
  fprintf(stBuilder,"\n");

  overlap = (int)ceil(((double)tr->mxtips * perCent) / adef->sets);
  
  fprintf(stderr," %d species will be included from each set for overlap\n", overlap); 

  while(v > 2)
    {
      v = v / 2;
      descent++;
    }
  
  
  
  stBuilderTree = (char *)malloc(treeStringLength);  
  count = (int *)malloc(sizeof(int) * numberOfSets);
  splittingNodes = (int *)malloc(sizeof(int) * numberOfSets);


  for(i = 1; i < (2 * tr->mxtips - 1); i++)
    tr->nodep[i]->set = 0;
  
  for(i = 0; i < numberOfSets; i++)
    count[i] = 0;

  centerSplit = findDivisorDistance(tr, 0);

  divideRec(tr, 0, descent, splittingNodes);
  for(i = 1; i <= tr->mxtips; i++)
    {
      /*fprintf(stderr,"%s %d\n",tr->nodep[i]->name, tr->nodep[i]->set);*/
      count[tr->nodep[i]->set] += 1;
    }

  /*for(i = 0; i < numberOfSets; i++)
    fprintf(stderr,"Count %d Divisor %d\n", count[i], splittingNodes[i]);*/

  filters = (int **)malloc(sizeof(int *) * numberOfSets);

  for(i = 0; i < numberOfSets; i++)
    {
      filters[i] = (int *)malloc((tr->mxtips + 1) * sizeof(int));
    }

  addSequencesSetAlgo2(tr, overlap, filters, count, centerSplit, numberOfSets);


  /*for(i = 0; i < numberOfSets; i++)
    {
      addSequencesSet(tr, i, overlap, splittingNodes[i], divideCount, filters[i], count); 
      }*/
  
#ifdef DEBUG
   {
    char str[10], checkpointFileName[1024], *divideTree, *outString;
    FILE *checkpointFile;
    divideTree = (char *) malloc(treeStringLength);

    strcpy(checkpointFileName, "RAxML_divide_");
    sprintf(str, "%d", divideCount);
    strcat(checkpointFileName, str);

    checkpointFile = fopen(checkpointFileName, "w");
    treeStringDivide(divideTree, tr, tr->start->back, 1);
    /*fprintf(stderr, "%s", divideTree);*/
    outString = divideTree;
    while(*outString++ != ']');
   
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(checkpointFile, "%c", *outString); 
	outString++;
      }	    
    fclose(checkpointFile);
    free(divideTree);
  }
#endif
  
  if(optimizeSubtrees)
    {
      names = (char **)malloc((tr->mxtips + 1) * sizeof(char *));
      for(i = 1; i <= tr->mxtips; i++)
	{
	  names[i] = tr->nodep[i]->name;
	}
  
      allTips = tr->mxtips;

      subtreeFile = fopen(subtreeFileName, "w");
      for(i = 0; i < numberOfSets; i++)
	{
	  RAxML_parsimony_Subtree_init(count[i], tr->rdta->sites, rdtaSupertree->y, names, filters[i],
 				       allTips);
	  RAxML_parsimony_Subtree(treeStringLength, &parsimonyTree); 
	  RAxML_parsimony_cleanup(); 	 
	  str_treeReadLenSpecial(parsimonyTree ,tr);
	  treeEvaluate(tr);
	  fprintf(stderr,"%f\n", tr->likelihood);
	  compute(tr, filters[i], adef);
#ifdef DEBUG
	  fprintf(stderr,"LH: %f\n", tr->likelihood);	  
#endif
	  
	  treeString(stBuilderTree, tr, tr->start->back, 1);
	  printTree(subtreeFile, tr, 0);
	  treeStringStBuilder(stBuilderTree, tr, tr->start->back, 1);
	  fprintf(stBuilder,"%s\n", stBuilderTree);
	  
	  free(filters[i]);
	}
      free(parsimonyTree);
      free(names);
      fclose(subtreeFile);
    }
 
      

  divideCount++;  
  fclose(stBuilder);
  /*check(tr, tr->nodep[splittingNodes[0]], 0);*/
  free(stBuilderTree);
  free(count);
  free(splittingNodes);
  free(filters);
}

void specialPrint(tree *tr)
{
  int i, j;
  char inverseMap[21] = {'z', 'A', 'C','M','G','R','S','V', 'T', 'W', 'Y','H','K','D','B','N','O','X','?','-'};
  FILE *outf; 
  outf = fopen("dummy", "w");

  fprintf(outf,"%d %d\n", tr->mxtips, tr->rdta->sites);

 for(i = 1; i <= tr->mxtips; i++)
    {     
      
      fprintf(outf, "T%d         \0", i);
	      /*fprintf(outf,"%s    ", tr->nodep[i]->name);*/
      for(j = 1; j <= tr->rdta->sites; j++)
	fprintf(outf, "%c", inverseMap[rdtaSupertree->y[i][j]]);
      fprintf(outf, "\n");	
    }
 fflush(outf);
 fclose(outf);
}
void printForStBuilder(tree *tr)
{
  FILE *stFile;
  int i;
  stFile = fopen("Build.input", "w");
  fprintf(stFile,"%d 3\n", tr->mxtips);
  fprintf(stFile,"*\n");
  for(i = 1; i <= tr->mxtips; i++)
    fprintf(stFile,"%d T%d *\n", i, i);
  fprintf(stFile,"%d\n", 4);
  fflush(stFile);
  fclose(stFile);

}
